/*
 *	NEWS - GTK-based RSS Ticker - Copyright (C) Emmanuel Thomas-Maurin 2009-2011
 *	<manutm007@gmail.com>
 *
 * 	This program is free software: you can redistribute it and/or modify
 * 	it under the terms of the GNU General Public License as published by
 * 	the Free Software Foundation, either version 3 of the License, or
 * 	(at your option) any later version.
 *
 * 	This program is distributed in the hope that it will be useful,
 * 	but WITHOUT ANY WARRANTY; without even the implied warranty of
 * 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * 	GNU General Public License for more details.
 *
 * 	You should have received a copy of the GNU General Public License
 * 	along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "news.h"

#ifdef G_OS_WIN32
extern FILE	*stdout_fp, *stderr_fp;
#endif

/*
 * render n lines of a stream (opened text file) into an image (pixmap)
 * of one long single line of text (n is limited because pixmap width must
 * be <= (PIXMAP_MAXLEN = 32 K - 1))
 * return adr of newly created pixmap
 *
 * *** one single stream is likely to be splitted into several text parts ***
 *
 * if returned pixmap is NULL then render_exit_status is set accordingly and
 * should be checked to know which error occured
 * (otherwise render_exit_status = OK or RENDER_NO_RESOURCE or
 * RENDER_LINE_TOO_LONG)
 * LINE_MAXLEN should be set to ensure that we won't have too long lines
 */
GdkDrawable *render_stream_to_pixmap(FILE *fp, FeedLinkAndOffset *link_and_offset, const Params *prm, \
		int *render_exit_status)
{
	GrEnv			*env = get_gr_env();
	char			*txt_str;
	PangoLayout		*p_layout;
	PangoFontDescription	*f_des;
	int			layout_width, layout_height;
	GdkGC			*gc_previous;
	GdkColor		shadow_color;
	guint16			shadow_color_alpha;
	float			shadow_k;
	char			tmp[4];
	int			i, j;

	*render_exit_status = OK;
	/*
	 * create layout
	 */
	p_layout = pango_layout_new(gtk_widget_get_pango_context(env->win));
	pango_layout_set_attributes(p_layout, NULL);
	pango_layout_set_single_paragraph_mode(p_layout, TRUE);

	f_des = pango_font_description_from_string((const char *)prm->font_n_s);
	pango_layout_set_font_description(p_layout, f_des);
	pango_font_description_free(f_des);
	pango_layout_context_changed(p_layout);
	/*
	 * get text string from stream
	 */
	txt_str = l_str_new(NULL);

	if (fp != NULL) {
		/* is text always utf-8 encoded -> ???? */
		txt_str = stream_to_htext(fp, txt_str, p_layout, prm, render_exit_status);
		if (*render_exit_status != OK && *render_exit_status != RENDER_LINE_TOO_LONG) {
			l_str_free(txt_str);
			return NULL;
		}
	} else {
		txt_str = void_stream_to_htext(txt_str, p_layout);
		*render_exit_status = RENDER_NO_RESOURCE;
	}
	/*
	 * translate link ranks inside (eventually splitted) text into offsets in pixmap
	 *
	 * scan text for "<LINKTAG_CHAR>00n" and get
	 * offset = layout_width(< text from start to tag >),
	 * for every rss item found
	 * then remove tags and fill an array of FeedLinkAndOffset's
	 * with offsets (rank and url variables are already filled)
	 *
	 */
	for (i = 0; i < NFEEDLINKANDOFFSETMAX; i++)	/* we reset all offsets to 0 */
		(link_and_offset + i)->offset_in_pixmap = 0;
	for (i = 0; txt_str[i] != '\0'; i++) {
		if (txt_str[i] == LINKTAG_CHAR) {
			str_n_cpy(tmp ,txt_str + i + 1, 3);
			j = atoi(tmp);
			txt_str[i] = '\0';
			if (j > 0 && j < NFEEDLINKANDOFFSETMAX)
				(link_and_offset + j)->offset_in_pixmap = get_layout_width(p_layout, txt_str);
			txt_str = l_str_cat(txt_str, txt_str + i + 4);
		}
	}
	/*
	 * fill layout
	 */
	/* TODO: could UTF-8 be broken here or somewhere in this func? */
	pango_layout_set_text(p_layout, txt_str, -1);
	l_str_free(txt_str);

	pango_layout_context_changed(p_layout);
	pango_layout_get_pixel_size2(p_layout, &layout_width, &layout_height);
	env->pixm_width = (gint)layout_width;
	env->pixm_height = (gint)layout_height;
	if (env->pixm_width > PIXMAP_MAXWIDTH) {
		*render_exit_status = RENDER_PIXMAP_TOO_WIDE;
		return NULL;
	}
	/*
	 * create pixmap
	 */
	env->pixmap = gdk_pixmap_new(NULL, env->pixm_width, env->pixm_height, env->depth);
	/*
	 * set gc and colormap
	 */
	gc_previous = env->gc;
	env->gc = gdk_gc_new(env->pixmap);
	if (GDK_IS_COLORMAP(env->colormap))
		gdk_gc_set_colormap(env->gc, env->colormap);
	if (gc_previous != NULL && GDK_IS_GC(gc_previous))
		g_object_unref(gc_previous);
	pango_layout_context_changed(p_layout);
	/*
	 * render layout to pixmap
	 * at the moment, alpha channel is not used (if depth = 32) because
	 * - it works with pixbuffers, not pixmaps, as far as I know
	 * - transparency can be set as a parameter for the top window
	 *   (for both text and background)
	 * well, could this be improved?
	 */
	if (prm->shadow == 'y') {
		/*
		 * compute shadow color
		 */
		if (prm->shadow_fx < 0)
			shadow_k = 1.0;
		else if (prm->shadow_fx > 10)
			shadow_k = 0.0;
		else
			shadow_k = 1.0 - (float)prm->shadow_fx / 10.0;

		shadow_color.red = (guint16)((float)prm->bg_color.red * shadow_k);
		shadow_color.green = (guint16)((float)prm->bg_color.green * shadow_k);
		shadow_color.blue = (guint16)((float)prm->bg_color.blue * shadow_k);
		shadow_color_alpha = prm->bg_color_alpha;
		/*
		 * create shadow by dual rendering with then without an offset
		 */
		gdk_draw_layout_with_colors(env->pixmap, env->gc, 0, 0, p_layout,
							NULL, &prm->bg_color);
		gdk_draw_layout_with_colors(env->pixmap, env->gc,
			prm->shadow_offset_x, prm->shadow_offset_y, p_layout,
						&shadow_color, &prm->bg_color);
		gdk_draw_layout_with_colors(env->pixmap, env->gc, 0, 0, p_layout,
							&prm->fg_color, NULL);
	} else
		gdk_draw_layout_with_colors(env->pixmap, env->gc, 0, 0, p_layout,
						&prm->fg_color, &prm->bg_color);

	if (p_layout != NULL)
		g_object_unref(p_layout);
	return env->pixmap;
}

/*
 * split stream into a a set of long lines so that, after rendering of each one,
 * we get a pixmap whose width is <= (PIXMAP_MAXLEN = 32 K - 1)
 * if one line is longer than tmp2_size, it's truncated and CUTLINE_DELIMITER is
 * appended but we should make sure that this won't happen
 */
char *stream_to_htext(FILE *fp, char *txt_str, PangoLayout *p_layout, const Params *prm, int *render_exit_status)
{
	GrEnv			*env = get_gr_env();
	static int		txt_splitted = FALSE;
	static char		*ending_str = NULL;
	long int		file_pos;
	char			*tmp1, *tmp2;
	size_t			tmp1_size = LINE_MAXLEN * 4;	/* anyways, this is adjusted by getline() */
	size_t			tmp2_size =  64 * 1024;		/* this should always be enough */
	int			c = 0;
	char			*empty_line;
	int			empty_line_length, empty_line_width, txt_str_width;/* len in char, width in pixels */
	int			i, j, k;
#ifndef G_OS_WIN32
	ssize_t			n;
#endif
	*render_exit_status = OK;
	/*
	 * creating an "empty line"
	 */
	empty_line = l_str_new(NULL);
	do
		empty_line = l_str_cat(empty_line, " ");
	while ((empty_line_width = get_layout_width(p_layout, empty_line)) < env->drwa_width);
	empty_line_length = strlen(empty_line);

	if (env->reload_rq)
		txt_splitted = FALSE;
	/*
	 * we start with an "empty line" but if text has been splitted
	 */
	if (!txt_splitted) {
		txt_str = l_str_cat(txt_str, empty_line);
		txt_str_width = empty_line_width;
	} else {
		if (ending_str != NULL) {
			txt_str = l_str_cat(txt_str, ending_str);
			txt_str_width = get_layout_width(p_layout, txt_str);
		} else {
			*render_exit_status = RENDER_NULL_ENDINGSTR;
			return txt_str;
		}
	}
	if (ending_str != NULL) {
		l_str_free(ending_str);
		ending_str = NULL;
	}
	txt_splitted = FALSE;
	/*
	 * then we read stream (text file) into txt_str one line at a time
	 * text file has previously been formatted with extra newline's so that
	 * no line is longer than LINE_MAXLEN
	 * we must make sure that no line needs to be cut (width > (PIXMAP_MAXLEN = 32 K - 1)
	 * after rendering) but how do we set LINE_MAXLEN properly?
	 * LINE_MAXLEN < (PIXMAP_MAXLEN = 32 K - 1) / FONT_MAXSIZE = about 163 if maxsize = 200
	 * so 128 should be always ok (see news.h)
	 *
	 * ============================================================================
	 * it would be much better, for performance reasons, to replace
	 * LINE_MAXLEN by a variable like: line_maxlen = k * PIXMAP_MAXLEN / font_size
	 * especially when font_size is big
	 * ============================================================================
	 */
	tmp1 = malloc2((tmp1_size + 1) * sizeof(char));	/* tmp1 will get 1 line */
	tmp2 = malloc2((tmp2_size + 1) * sizeof(char));	/* tmp2 will get 1 "expanded" line */
	/*
	 * txt_str will get all n lines
	 */
	while (1) {
		file_pos = ftell(fp);
#ifndef G_OS_WIN32
		if ((n = getline(&tmp1, &tmp1_size, fp)) == -1) {
#else
		if (fgets(tmp1, tmp1_size, fp) == NULL) {
#endif
			/*
			 * eof or error?
			 */
			if (feof(fp) == 0) {
				*render_exit_status = RENDER_GETLINE_ERROR;
				return txt_str;
			}
			/*
			 * text file fully read -> reset to beginning
			 */
			fseek(fp, 0, SEEK_SET);
			break;
		} else {
#ifndef G_OS_WIN32
			for (i = 0, j = 0; i < n; i++) {
#else
			for (i = 0, j = 0; i < (int)strlen(tmp1); i++) {
#endif
				/*
				 * this is not supposed to happen but if tmp2_size was wrongly set
				 */
				if ((size_t)j >= tmp2_size - strlen(prm->cutline_delimiter) - 1) {
					str_n_cpy(&tmp2[j], prm->cutline_delimiter, OPTION_VALUE_MAXLEN);
					j += strlen(prm->cutline_delimiter);
					warning("stream_to_htext():", "line too long", "", "");
					*render_exit_status = RENDER_LINE_TOO_LONG;
					break;
				}
				c = tmp1[i];
				k = 1;

				if (c == '\n') {
					str_n_cpy(&tmp2[j], prm->line_delimiter, OPTION_VALUE_MAXLEN);
					j += strlen(prm->line_delimiter);
					break;
				}
				if (prm->special_chars == 'y') {
					if (c == prm->new_page_char) {
						c = ' ';
						k = empty_line_length;
					} else if (c == prm->tab_char) {
						c = ' ';
						k = TAB_SIZE;
					}
				}
				while (k-- > 0)
					tmp2[j++] = c;
			}
			tmp2[j] = '\0';
		}
		txt_str_width += get_layout_width(p_layout, tmp2);

		if (txt_str_width > PIXMAP_MAXWIDTH - 2 * empty_line_width - prm->shadow_offset_x) {
			/*
			 * need to split text into several parts
			 * we don't add this line but will use it in next pixmap
			 */
			fseek(fp, file_pos, SEEK_SET);
			txt_splitted = TRUE;
			break;
		} else
			txt_str = l_str_cat(txt_str, tmp2);
	}
	/*
	 * and we end with an "empty line" but if text has been splitted
	 */
	if (!txt_splitted) {
		txt_str = l_str_cat(txt_str, empty_line);
		ending_str = NULL;
	} else {
		/*
		 * in which case we keep "tail" of txt_str whose width is drwa_width
		 */
		i = strlen(txt_str) - 1;
		while (get_layout_width(p_layout, txt_str + i--) <= env->drwa_width);
		ending_str = l_str_new(txt_str + i + 1);
	}
	free2(tmp1);
	free2(tmp2);
	l_str_free(empty_line);

	return txt_str;
}

char *void_stream_to_htext(char *txt_str, PangoLayout *p_layout)
{
	GrEnv	*env = get_gr_env();
	char	*empty_line;

	empty_line = l_str_new(NULL);
	do
		empty_line = l_str_cat(empty_line, " ");
	while (get_layout_width(p_layout, empty_line) < env->drwa_width);

	txt_str = l_str_cat(txt_str, empty_line);
	txt_str = l_str_cat(txt_str, "No resource");
	txt_str = l_str_cat(txt_str, empty_line);
	l_str_free(empty_line);

	return txt_str;
}

int get_fsize_from_layout_height(gint height, const char *fontname)
{
	PangoLayout		*p_layout;
	PangoFontDescription	*f_des;
	char			fnamesize[FONT_MAXLEN + 1];
	char			fname[FONT_NAME_MAXLEN + 1], fsize[FONT_SIZE_MAXLEN + 1];
	char			check_str[] = "gjpqyÀÄÅÇÉÖÜ€£$!&({[?|";	/* should be enough - is it? */
	gint			height2;
	int			i = 1;

	p_layout = pango_layout_new(gtk_widget_get_pango_context(get_gr_env()->win));
	pango_layout_set_attributes(p_layout, NULL);
	pango_layout_set_single_paragraph_mode(p_layout, TRUE);

	str_n_cpy(fname, fontname, FONT_NAME_MAXLEN);

	do {
		snprintf(fsize, FONT_SIZE_MAXLEN + 1, "%3d", i++);
		compact_font(fnamesize, fname, fsize);
		f_des = pango_font_description_from_string((const char *)fnamesize);
		pango_layout_set_font_description(p_layout, f_des);
		pango_font_description_free(f_des);
		height2 = get_layout_height(p_layout, check_str);
	} while (height2 < height && i < 1000);

	if (p_layout != NULL)
		g_object_unref(p_layout);

	return (i - 1);
}

gint get_layout_height_from_fnamesize(const char *fnamesize)
{
	PangoLayout		*p_layout;
	PangoFontDescription	*f_des;
	char			check_str[] = "gjpqyÀÄÅÇÉÖÜ€£$!&({[?|";
	gint			height;

	p_layout = pango_layout_new(gtk_widget_get_pango_context(get_gr_env()->win));
	pango_layout_set_attributes(p_layout, NULL);
	pango_layout_set_single_paragraph_mode(p_layout, TRUE);

	f_des = pango_font_description_from_string(fnamesize);
	pango_layout_set_font_description(p_layout, f_des);
	pango_font_description_free(f_des);

	height = get_layout_height(p_layout, check_str);
	if (p_layout != NULL)
		g_object_unref(p_layout);

	return height;
}

gint get_layout_width(PangoLayout *p_layout, const char *str)
{
	int	layout_width, layout_height;

	pango_layout_set_text(p_layout, str, -1);
	pango_layout_context_changed(p_layout);
	pango_layout_get_pixel_size2(p_layout, &layout_width, &layout_height);

	return (gint)layout_width;
}

gint get_layout_height(PangoLayout *p_layout, const char *str)
{
	int	layout_width, layout_height;

	pango_layout_set_text(p_layout, str, -1);
	pango_layout_context_changed(p_layout);
	pango_layout_get_pixel_size2(p_layout, &layout_width, &layout_height);

	return (gint)layout_height;
}

/* this avoid troubles in case of an empty string inside layout */
void pango_layout_get_pixel_size2(PangoLayout *p_layout, int *layout_width, int *layout_height)
{
	if (strlen(pango_layout_get_text(p_layout)) > 0)
		pango_layout_get_pixel_size(p_layout, layout_width, layout_height);
	else {
		*layout_width = 0;
		*layout_height = 0;
		/*big_error(RENDER_OTHER_ERROR,
			"news_misc.c:", "pango_layout_get_pixel_size2():", "Text length = 0", "");*/
		fprintf(STD_ERR,
			"news_misc.c: pango_layout_get_pixel_size2(): Text length = 0 - Troubles ahead ?\n");
	}
	return;
}
